# -*- coding: utf-8 -*-
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Firefox IndexedDB records."""
import pathlib
import sqlite3
import sys
import traceback
from dataclasses import dataclass
from typing import Any, Generator, List, Optional

from dfindexeddb import errors
from dfindexeddb.indexeddb.firefox import gecko


@dataclass
class FirefoxBlobInfo:
  """Firefox IndexedDB blob info.

  Attributes:
    file_id: the file identifier.
    file_path: the full path to the blob file.
    blob_data: the blob data.
  """

  file_id: str
  file_path: str
  blob_data: Any = None


@dataclass
class FirefoxObjectStoreInfo:
  """A FireFox ObjectStoreInfo.

  Attributes:
    id: the object store ID.
    name: the object store name.
    key_path: the object store key path.
    auto_inc: the current auto-increment value.
      database_name: the database name from the database table.
  """

  id: int
  name: str
  key_path: str
  auto_inc: int
  database_name: str


@dataclass
class FirefoxIndexedDBRecord:
  """A Firefox IndexedDBRecord.

  Attributes:
    key: the parsed key.
    value: the parsed value.
    file_ids: the file identifiers.
    object_store_id: the object store id.
    object_store_name: the object store name from the object_store table.
    database_name: the IndexedDB database name from the database table.
    raw_key: the raw key.
    raw_value: the raw value.
  """

  key: Any
  value: Any
  file_ids: Optional[str]
  object_store_id: int
  object_store_name: str
  database_name: str
  raw_key: Optional[bytes] = None
  raw_value: Optional[bytes] = None
  blobs: Optional[List[FirefoxBlobInfo]] = None

  @property
  def is_key_filterable(self) -> bool:
    """True if the record key is filterable."""
    return True

  @property
  def is_value_filterable(self) -> bool:
    """True if the record value is filterable."""
    return True

  def MatchesKey(self, term: str) -> bool:
    """Returns True if the record key matches the filter term.

    Args:
      term: the filter term.
    """
    key_val = getattr(self.key, "value", self.key)
    return term in str(key_val)

  def MatchesValue(self, term: str) -> bool:
    """Returns True if the record value or blobs matches the filter term.

    Args:
      term: the filter term.
    """
    if term in str(self.value):
      return True

    if self.blobs:
      for blob_info in self.blobs:
        if blob_info.blob_data is not None and term in str(blob_info.blob_data):
          return True
    return False


class FileReader:
  """A reader for Firefox IndexedDB sqlite3 files.

  Attributes:
    database_name: the database name.
    origin: the database origin.
    metadata_version: the metadata version.
    last_vacuum_time: the last vacuum time.
    last_analyze_time: the last analyze time.
  """

  def __init__(self, filename: str):
    """Initializes the FileReader.

    Args:
      filename: the IndexedDB filename.
    """
    self.filename = filename
    self._uri = pathlib.Path(filename).resolve().as_uri()

    with sqlite3.connect(f"{self._uri}?mode=ro", uri=True) as conn:
      cursor = conn.execute(
          "SELECT name, origin, version, last_vacuum_time, last_analyze_time "
          "FROM database"
      )
      result = cursor.fetchone()
      self.database_name = result[0]
      self.origin = result[1]
      self.metadata_version = result[2]
      self.last_vacuum_time = result[3]
      self.last_analyze_time = result[4]

  def _ParseKey(self, key: bytes) -> Any:
    """Parses a key."""
    try:
      return gecko.IDBKey.FromBytes(key)
    except errors.ParserError as e:
      print("failed to parse", key, file=sys.stderr)
      traceback.print_exception(type(e), e, e.__traceback__)
      return key

  def _ParseValue(self, value: bytes) -> Any:
    """Parses a value."""
    try:
      return gecko.JSStructuredCloneDecoder.FromBytes(value)
    except errors.ParserError as err:
      print("failed to parse", value, file=sys.stderr)
      traceback.print_exception(type(err), err, err.__traceback__)
      return value

  def LoadBlobsForRecord(
      self, file_ids: Optional[str] = None
  ) -> List[FirefoxBlobInfo]:
    """Returns the FirefoxBlobInfo instances for the given file_ids string.

    Args:
      file_ids: the file identifiers (e.g. ".2" or "1 2 3").

    Returns:
      a list of FirefoxBlobInfo instances.
    """
    blobs: list[FirefoxBlobInfo] = []
    if not file_ids:
      return blobs

    ids = file_ids.strip(".").split(" ")
    for file_id in ids:
      if not file_id:
        continue
      file_name = str(abs(int(file_id)))  # mutable file ids are -ve?
      blob_file_path = (
          pathlib.Path(self.filename).with_suffix(".files") / file_name
      )
      blob_data = None
      if blob_file_path.exists():
        blob_data = blob_file_path.read_bytes()
        try:
          blob_data = gecko.JSStructuredCloneDecoder.FromBytes(blob_data)
        except errors.ParserError as err:
          print(
              f"Failed to parse blob {file_id}, defaulting to raw value as it "
              f"is unlikely to be a JS value: {err}",
              file=sys.stderr,
          )

      blobs.append(
          FirefoxBlobInfo(
              file_id=file_id,
              file_path=str(blob_file_path),
              blob_data=blob_data,
          )
      )
    return blobs

  def ObjectStores(self) -> Generator[FirefoxObjectStoreInfo, None, None]:
    """Returns the Object Store information from the IndexedDB database.

    Yields:
      FirefoxObjectStoreInfo instances.
    """
    with sqlite3.connect(f"{self._uri}?mode=ro", uri=True) as conn:
      cursor = conn.execute(
          "SELECT id, auto_increment, name, key_path FROM object_store"
      )
      results = cursor.fetchall()
      for result in results:
        yield FirefoxObjectStoreInfo(
            id=result[0],
            name=result[2],
            key_path=result[3],
            auto_inc=result[1],
            database_name=self.database_name,
        )

  def _EnumerateCursor(
      self,
      cursor: sqlite3.Cursor,
      include_raw_data: bool = False,
      load_blobs: bool = True,
  ) -> Generator[FirefoxIndexedDBRecord, None, None]:
    """Yields FirefoxIndexedDBRecord records from a sqlite3 cursor.

    Args:
      cursor: the sqlite3 cursor.
      include_raw_data: whether to include the raw data.
      load_blobs: whether to load the record blobs.

    Yields:
      FirefoxIndexedDBRecord records.
    """
    for row in cursor:
      key = self._ParseKey(row[0])
      file_ids = row[3]
      if isinstance(file_ids, bytes):
        file_ids = file_ids.decode("utf-8")

      if isinstance(row[1], bytes):
        value = self._ParseValue(row[1])
      else:
        value = row[1]

      blobs = None
      if load_blobs and file_ids:
        blobs = self.LoadBlobsForRecord(file_ids)

      yield FirefoxIndexedDBRecord(
          key=key,
          value=value,
          object_store_id=row[2],
          file_ids=file_ids,
          object_store_name=row[4].decode("utf-8"),
          database_name=self.database_name,
          raw_key=row[0] if include_raw_data else None,
          raw_value=row[1] if include_raw_data else None,
          blobs=blobs,
      )

  def RecordsByObjectStoreId(
      self,
      object_store_id: int,
      include_raw_data: bool = False,
      load_blobs: bool = True,
  ) -> Generator[FirefoxIndexedDBRecord, None, None]:
    """Returns FirefoxIndexedDBRecords by a given object store id.

    Args:
      object_store_id: the object store id.
      include_raw_data: whether to include the raw data.
      load_blobs: whether to load the record blobs.
    """
    with sqlite3.connect(f"{self._uri}?mode=ro", uri=True) as conn:
      conn.text_factory = bytes
      cursor = conn.execute(
          "SELECT od.key, od.data, od.object_store_id, od.file_ids, os.name "
          "FROM object_data od "
          "JOIN object_store os ON od.object_store_id == os.id "
          "WHERE os.id = ? ORDER BY od.key",
          (object_store_id,),
      )
      yield from self._EnumerateCursor(cursor, include_raw_data, load_blobs)

  def Records(
      self, include_raw_data: bool = False, load_blobs: bool = True
  ) -> Generator[FirefoxIndexedDBRecord, None, None]:
    """Returns FirefoxIndexedDBRecords from the database

    Args:
      include_raw_data: whether to include the raw data.
      load_blobs: whether to load the record blobs.

    Yields:
      FirefoxIndexedDBRecord instances.
    """
    with sqlite3.connect(f"{self._uri}?mode=ro", uri=True) as conn:
      conn.text_factory = bytes
      cursor = conn.execute(
          "SELECT od.key, od.data, od.object_store_id, od.file_ids, os.name "
          "FROM object_data od "
          "JOIN object_store os ON od.object_store_id == os.id"
      )
      yield from self._EnumerateCursor(cursor, include_raw_data, load_blobs)


class FolderReader:
  """A reader for a FireFox IndexedDB folder.

  The path takes a general form of ./<origin>/idb/<filename>.[files|sqlite]

  """

  def __init__(self, folder_name: pathlib.Path):
    """Initializes the FireFox IndexedDB FolderReader.

    Args:
      folder_name: the IndexedDB folder name (the origin folder).

    Raises:
      ValueError: if the folder does not exist or is not a directory.
    """
    self.folder_name = folder_name
    if not self.folder_name.exists():
      raise ValueError(f"{folder_name} does not exist.")
    if not self.folder_name.is_dir():
      raise ValueError(f"{folder_name} is not a directory.")

    self.file_names: list[pathlib.Path] = []
    for file_name in self.folder_name.rglob("idb/*.sqlite"):
      self.file_names.append(file_name)

  def Records(
      self, include_raw_data: bool = False, load_blobs: bool = True
  ) -> Generator[FirefoxIndexedDBRecord, None, None]:
    """Returns FirefoxIndexedDBRecords from the IndexedDB folder."""
    for file_name in self.file_names:
      yield from FileReader(str(file_name)).Records(
          include_raw_data=include_raw_data, load_blobs=load_blobs
      )



# -*- coding: utf-8 -*-
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A CLI tool for dfindexeddb."""
import argparse
import dataclasses
import enum
import json
import pathlib
from datetime import datetime
from typing import Any

from dfindexeddb import utils, version
from dfindexeddb.indexeddb import types
from dfindexeddb.indexeddb.chromium import blink
from dfindexeddb.indexeddb.chromium import sqlite
from dfindexeddb.indexeddb.chromium import record as chromium_record
from dfindexeddb.indexeddb.firefox import gecko
from dfindexeddb.indexeddb.firefox import record as firefox_record
from dfindexeddb.indexeddb.safari import record as safari_record

_VALID_PRINTABLE_CHARACTERS = (
    " abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"
    + "!\"#$%&'()*+,-./:;<=>?@[\\]^_`{|}~."
)


class Encoder(json.JSONEncoder):
  """A JSON encoder class for dfindexeddb fields."""

  def default(self, o):  # type: ignore[no-untyped-def]
    if dataclasses.is_dataclass(o):
      o_dict = utils.asdict(o)
      return o_dict
    if isinstance(o, (bytes, bytearray)):
      out = []
      for x in o:
        if chr(x) not in _VALID_PRINTABLE_CHARACTERS:
          out.append(f"\\x{x:02X}")
        else:
          out.append(chr(x))
      return "".join(out)
    if isinstance(o, datetime):
      return o.isoformat()
    if isinstance(o, types.Undefined):
      return "<undefined>"
    if isinstance(o, types.JSArray):
      return o.__dict__
    if isinstance(o, types.Null):
      return "<null>"
    if isinstance(o, set):
      return list(o)
    if isinstance(o, types.RegExp):
      return str(o)
    if isinstance(o, enum.Enum):
      return o.name
    return json.JSONEncoder.default(self, o)


def _Output(structure: Any, output: str) -> None:
  """Helper method to output parsed structure to stdout.

  Args:
    structure: The structure to output.
    output: The output format.
  """
  if output == "json":
    print(json.dumps(structure, indent=2, cls=Encoder))
  elif output == "jsonl":
    print(json.dumps(structure, cls=Encoder))
  elif output == "repr":
    print(structure)


def BlinkCommand(args: argparse.Namespace) -> None:
  """The CLI for processing a file as a blink-encoded value.

  Args:
    args: The arguments for processing the blink-encoded value.
  """
  with open(args.source, "rb") as fd:
    buffer = fd.read()
    blink_value = blink.V8ScriptValueDecoder.FromBytes(buffer)
    _Output(blink_value, output=args.output)


def GeckoCommand(args: argparse.Namespace) -> None:
  """The CLI for processing a file as a gecko-encoded value.

  Args:
    args: The arguments for processing the gecko-encoded value.
  """
  with open(args.source, "rb") as fd:
    buffer = fd.read()
    blink_value = gecko.JSStructuredCloneDecoder.FromBytes(buffer)
    _Output(blink_value, output=args.output)


def _MatchesFilters(
    record: types.FilterableRecord, args: argparse.Namespace
) -> bool:
  """Returns True if the record matches the filter criteria.

  Supported filters:
  * object_store_id - filters by object store ID
  * filter_key - filters by key
  * filter_value - filters by value

  Args:
    record: The record to check for filtering.
    args: The arguments containing filter criteria.

  Returns:
    True if the record matches the filter criteria or no filters are set, False
        otherwise.
  """
  if (
      args.object_store_id is not None
      and record.object_store_id != args.object_store_id
  ):
    return False

  if args.filter_value is not None:
    if not record.is_value_filterable:
      return False
    if not record.MatchesValue(args.filter_value):
      return False

  if args.filter_key is not None:
    if not record.is_key_filterable:
      return False
    if not record.MatchesKey(args.filter_key):
      return False

  return True


def HandleChromiumDB(args: argparse.Namespace) -> None:
  """The handler for processing a directory/file as Chromium IndexedDB.

  If the source is a directory, it will process as a LevelDB based IndexedDB.
  If the source is a file, it will process as a SQLite based IndexedDB.

  Args:
    args: The arguments for processing the Chromium IndexedDB.
  """
  if args.source.is_file():
    reader = sqlite.DatabaseReader(str(args.source))
    if args.object_store_id is not None:
      records: Any = reader.RecordsByObjectStoreId(
          args.object_store_id,
          include_raw_data=args.include_raw_data,
          load_blobs=args.load_blobs,
      )
    else:
      records = reader.Records(
          include_raw_data=args.include_raw_data,
          load_blobs=args.load_blobs,
      )
  else:
    records = chromium_record.FolderReader(args.source).GetRecords(
        use_manifest=args.use_manifest,
        use_sequence_number=args.use_sequence_number,
        include_raw_data=args.include_raw_data,
        load_blobs=args.load_blobs,
    )

  for record in records:
    if _MatchesFilters(record, args):
      _Output(record, output=args.output)


def HandleFirefoxDB(args: argparse.Namespace) -> None:
  """The handler for processing a file as Firefox IndexedDB.

  Args:
    args: The arguments for processing the Firefox IndexedDB.
  """
  reader = firefox_record.FileReader(str(args.source))
  if args.object_store_id is not None:
    records = reader.RecordsByObjectStoreId(
        args.object_store_id,
        include_raw_data=args.include_raw_data,
        load_blobs=args.load_blobs,
    )
  else:
    records = reader.Records(
        include_raw_data=args.include_raw_data,
        load_blobs=args.load_blobs,
    )

  for record in records:
    if _MatchesFilters(record, args):
      _Output(record, output=args.output)


def HandleSafariDB(args: argparse.Namespace) -> None:
  """The handler for processing a file as Safari IndexedDB.

  Args:
    args: The arguments for processing the Safari IndexedDB.
  """
  reader = safari_record.FileReader(str(args.source))
  if args.object_store_id is not None:
    records = reader.RecordsByObjectStoreId(
        args.object_store_id,
        include_raw_data=args.include_raw_data,
        load_blobs=args.load_blobs,
    )
  else:
    records = reader.Records(
        include_raw_data=args.include_raw_data,
        load_blobs=args.load_blobs,
    )

  for record in records:
    if _MatchesFilters(record, args):
      _Output(record, output=args.output)


def DbCommand(args: argparse.Namespace) -> None:
  """The CLI for processing a file/directory as IndexedDB.

  Args:
    args: The arguments for processing the IndexedDB.
  """
  if args.format in ("chrome", "chromium"):
    HandleChromiumDB(args)
  elif args.format == "firefox":
    HandleFirefoxDB(args)
  elif args.format == "safari":
    HandleSafariDB(args)


def LdbCommand(args: argparse.Namespace) -> None:
  """The CLI for processing a LevelDB table (.ldb) file as IndexedDB.

  Args:
    args: The arguments for processing the LevelDB table.
  """
  for db_record in chromium_record.ChromiumIndexedDBRecord.FromFile(
      args.source,
      include_raw_data=args.include_raw_data,
      load_blobs=args.load_blobs,
  ):
    if _MatchesFilters(db_record, args):
      _Output(db_record, output=args.output)


def LogCommand(args: argparse.Namespace) -> None:
  """The CLI for processing a LevelDB log file as IndexedDB.

  Args:
    args: The arguments for processing the LevelDB log file.
  """
  for db_record in chromium_record.ChromiumIndexedDBRecord.FromFile(
      args.source,
      include_raw_data=args.include_raw_data,
      load_blobs=args.load_blobs,
  ):
    if _MatchesFilters(db_record, args):
      _Output(db_record, output=args.output)


def App() -> None:
  """The CLI app entrypoint for dfindexeddb."""
  parser = argparse.ArgumentParser(
      prog="dfindexeddb",
      description="A cli tool for parsing IndexedDB files",
      epilog=f"Version {version.GetVersion()}",
  )

  subparsers = parser.add_subparsers()

  parser_blink = subparsers.add_parser(
      "blink", help="Parse a file as a blink-encoded value."
  )
  parser_blink.add_argument(
      "-s",
      "--source",
      required=True,
      type=pathlib.Path,
      help="The source file.",
  )
  parser_blink.add_argument(
      "-o",
      "--output",
      choices=["json", "jsonl", "repr"],
      default="json",
      help="Output format.  Default is json.",
  )
  parser_blink.set_defaults(func=BlinkCommand)

  parser_gecko = subparsers.add_parser(
      "gecko", help="Parse a file as a gecko-encoded value."
  )
  parser_gecko.add_argument(
      "-s",
      "--source",
      required=True,
      type=pathlib.Path,
      help="The source file.",
  )
  parser_gecko.add_argument(
      "-o",
      "--output",
      choices=["json", "jsonl", "repr"],
      default="json",
      help="Output format.  Default is json.",
  )
  parser_gecko.set_defaults(func=GeckoCommand)

  parser_db = subparsers.add_parser(
      "db", help="Parse a directory/file as IndexedDB."
  )
  parser_db.add_argument(
      "-s",
      "--source",
      required=True,
      type=pathlib.Path,
      help=(
          "The source IndexedDB folder (for chrome/chromium) "
          "or sqlite3 file (for firefox/safari)."
      ),
  )
  recover_group = parser_db.add_mutually_exclusive_group()
  recover_group.add_argument(
      "--use_manifest",
      action="store_true",
      help="Use manifest file to determine active/deleted records.",
  )
  recover_group.add_argument(
      "--use_sequence_number",
      action="store_true",
      help=(
          "Use sequence number and file offset to determine active/deleted "
          "records."
      ),
  )
  parser_db.add_argument(
      "-f",
      "--format",
      required=True,
      choices=["chromium", "chrome", "firefox", "safari"],
      help="The type of IndexedDB to parse.",
  )
  parser_db.add_argument(
      "--object_store_id",
      type=int,
      help="The object store ID to filter by.",
  )
  parser_db.add_argument(
      "--include_raw_data",
      action="store_true",
      help="Include raw key and value bytes for each record in the output.",
  )
  parser_db.add_argument(
      "--load_blobs",
      action="store_true",
      default=False,
      help="Load blob data, if available for each record in the output.",
  )
  parser_db.add_argument(
      "-o",
      "--output",
      choices=["json", "jsonl", "repr"],
      default="json",
      help="Output format.  Default is json.",
  )
  parser_db.add_argument(
      "--filter_value",
      type=str,
      help=(
          "Only output records where the value contains this string. "
          "Values are normalized to strings before comparison."
      ),
  )
  parser_db.add_argument(
      "--filter_key",
      type=str,
      help=(
          "Only output records where the key contains this string. "
          "Keys are normalized to strings before comparison."
      ),
  )
  parser_db.set_defaults(func=DbCommand)

  parser_ldb = subparsers.add_parser(
      "ldb", help="Parse a ldb file as IndexedDB."
  )
  parser_ldb.add_argument(
      "-s",
      "--source",
      required=True,
      type=pathlib.Path,
      help="The source .ldb file.",
  )
  parser_ldb.add_argument(
      "-o",
      "--output",
      choices=["json", "jsonl", "repr"],
      default="json",
      help="Output format.  Default is json.",
  )
  parser_ldb.add_argument(
      "--object_store_id",
      type=int,
      help="The object store ID to filter by.",
  )
  parser_ldb.add_argument(
      "--include_raw_data",
      action="store_true",
      help="Include raw key and value bytes for each record in the output.",
  )
  parser_ldb.add_argument(
      "--load_blobs",
      action="store_true",
      default=False,
      help="Load blob data, if available for each record in the output.",
  )
  parser_ldb.add_argument(
      "--filter_value",
      type=str,
      help=(
          "Only output records where the value contains this string. "
          "Values are normalized to strings before comparison."
      ),
  )
  parser_ldb.add_argument(
      "--filter_key",
      type=str,
      help=(
          "Only output records where the key contains this string. "
          "Keys are normalized to strings before comparison."
      ),
  )
  parser_ldb.set_defaults(func=LdbCommand)

  parser_log = subparsers.add_parser(
      "log", help="Parse a log file as IndexedDB."
  )
  parser_log.add_argument(
      "-s",
      "--source",
      required=True,
      type=pathlib.Path,
      help="The source .log file.",
  )
  parser_log.add_argument(
      "-o",
      "--output",
      choices=["json", "jsonl", "repr"],
      default="json",
      help="Output format.  Default is json.",
  )
  parser_log.add_argument(
      "--object_store_id",
      type=int,
      help="The object store ID to filter by.",
  )
  parser_log.add_argument(
      "--include_raw_data",
      action="store_true",
      help="Include raw key and value bytes for each record in the output.",
  )
  parser_log.add_argument(
      "--load_blobs",
      action="store_true",
      default=False,
      help="Load blob data, if available for each record in the output.",
  )
  parser_log.add_argument(
      "--filter_value",
      type=str,
      help=(
          "Only output records where the value contains this string. "
          "Values are normalized to strings before comparison."
      ),
  )
  parser_log.add_argument(
      "--filter_key",
      type=str,
      help=(
          "Only output records where the key contains this string. "
          "Keys are normalized to strings before comparison."
      ),
  )
  parser_log.set_defaults(func=LogCommand)

  args: argparse.Namespace = parser.parse_args()
  if hasattr(args, "func"):
    args.func(args)
  else:
    parser.print_help()



# -*- coding: utf-8 -*-
# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Chromium IndexedDB records encoded in sqlite3 databases."""

import pathlib
import sqlite3
from typing import Any, Generator, Optional
from dataclasses import dataclass

import snappy
import zstd

from dfindexeddb.indexeddb.chromium import blink
from dfindexeddb.indexeddb.chromium import definitions
from dfindexeddb.indexeddb.chromium import record


@dataclass
class ChromiumIndexedDBRecord:
  """Chromium IndexedDB record parsed from sqlite3 database.

  Attributes:
    row_id: the row ID.
    object_store_id: the object store ID.
    compression_type: the compression type.
    key: the key.
    value: the value.
    has_blobs: whether the record has blobs.
    raw_key: the raw key.
    raw_value: the raw value.
  """

  row_id: int
  object_store_id: int
  compression_type: int
  key: Any
  value: Any
  has_blobs: bool
  raw_key: Optional[bytes]
  raw_value: Optional[bytes]

  @property
  def is_key_filterable(self) -> bool:
    """True if the record key is filterable."""
    return True

  @property
  def is_value_filterable(self) -> bool:
    """True if the record value is filterable."""
    return True

  def MatchesKey(self, term: str) -> bool:
    """Returns True if the record key matches the filter term.

    Args:
      term: the filter term.
    """
    key_val = getattr(self.key, "value", self.key)
    return term in str(key_val)

  def MatchesValue(self, term: str) -> bool:
    """Returns True if the record value matches the filter term.

    Args:
      term: the filter term.
    """
    if self.has_blobs and self.value:
      return any(term in str(blob_info.blob_data) for blob_info in self.value)
    return term in str(self.value)


@dataclass
class ChromiumObjectStoreInfo:
  """Chromium IndexedDB object store info parsed from sqlite3 database.

  Attributes:
    id: the object store ID.
    name: the object store name.
    key_path: the object store key path.
    auto_increment: whether the object store is auto increment.
    key_generator_current_number: the current number of the key generator.
  """

  id: int
  name: str
  key_path: str
  auto_increment: int
  key_generator_current_number: int


@dataclass
class ChromiumBlobInfo:
  """Chromium IndexedDB blob info parsed from sqlite3 database.

  Attributes:
    row_id: the blob row ID.
    object_type: the object type.
    mime_type: the mime type.
    size_bytes: the total size in bytes.
    file_name: the file name (only for files).
    number_of_chunks: the number of chunks including the initial one.
    blob_data: the blob data.
  """

  row_id: int
  object_type: int
  mime_type: Optional[str]
  size_bytes: int
  file_name: Optional[str]
  number_of_chunks: int
  blob_data: bytes


class DatabaseReader:
  """A reader for Chromium IndexedDB sqlite3 files.

  Attributes:
    filename: the path to the sqlite3 file.
  """

  def __init__(self, filename: str):
    """Initializes the reader.

    Args:
      filename: the path to the sqlite3 file.
    """
    self.filename = filename
    self._uri = pathlib.Path(filename).resolve().as_uri()

  def ObjectStores(self) -> Generator[ChromiumObjectStoreInfo, None, None]:
    """Yields object stores."""
    with sqlite3.connect(f"{self._uri}?mode=ro", uri=True) as conn:
      cursor = conn.cursor()
      cursor.execute(definitions.SQL_OBJECT_STORES_QUERY)
      for row in cursor:
        yield ChromiumObjectStoreInfo(
            id=row[0],
            name=row[1].decode("utf-16-le"),
            key_path=row[2],
            auto_increment=row[3],
            key_generator_current_number=row[4],
        )

  def _GetLegacyBlobPath(self, blob_id: int) -> str:
    """Gets the path to a legacy blob file.

    Args:
      blob_id: the blob ID.

    Returns:
      The path to the legacy blob file.
    """
    db_dir = pathlib.Path(self.filename).parent
    blob_path = db_dir / f"{blob_id:x}"
    return str(blob_path)

  def LoadLegacyBlobData(self, blob_id: int) -> bytes:
    """Loads legacy blob data from disk.

    Args:
      blob_id: the blob ID.

    Returns:
      The blob data.

    Raises:
      FileNotFoundError: if the legacy blob file is not found.
    """
    blob_path = pathlib.Path(self._GetLegacyBlobPath(blob_id))
    if blob_path.exists():
      return blob_path.read_bytes()
    raise FileNotFoundError(f"Legacy blob file not found: {blob_path}")

  def LoadBlobDataForRecordId(
      self, row_id: int
  ) -> Generator[ChromiumBlobInfo, None, None]:
    """Loads blob data for a given record row ID.

    Args:
      row_id: the record row ID.

    Yields:
      ChromiumBlobInfo objects.
    """
    with sqlite3.connect(f"{self._uri}?mode=ro", uri=True) as conn:
      conn.row_factory = sqlite3.Row
      cursor = conn.cursor()

      # Note this is a UNION query between the blob and overflow_blob_chunks
      # table.  The chunk_index = 0 for the row from the 'blobs' table.
      cursor.execute(definitions.SQL_BLOB_DATA_QUERY, (row_id, row_id))

      current_blob_id = None
      current_blob_data = bytearray()
      current_record: Optional[sqlite3.Row] = None
      total_number_of_chunks = 0

      for blob_row in cursor:
        blob_id = blob_row["row_id"]

        if blob_id != current_blob_id:
          if current_record is not None:
            yield ChromiumBlobInfo(
                row_id=current_record["row_id"],
                object_type=current_record["object_type"],
                mime_type=current_record["mime_type"],
                size_bytes=current_record["size_bytes"],
                file_name=current_record["file_name"],
                number_of_chunks=total_number_of_chunks,
                blob_data=bytes(current_blob_data),
            )
          current_blob_id = blob_id
          current_blob_data = bytearray()
          current_record = blob_row
          total_number_of_chunks = 0

        if blob_row["chunk_index"] == 0 and blob_row["bytes"] is None:
          current_blob_data.extend(self.LoadLegacyBlobData(blob_id))
          total_number_of_chunks += 1
          continue

        if blob_row["bytes"]:
          current_blob_data.extend(blob_row["bytes"])
          total_number_of_chunks += 1

      if current_record is not None:
        yield ChromiumBlobInfo(
            row_id=current_record["row_id"],
            object_type=current_record["object_type"],
            mime_type=current_record["mime_type"],
            size_bytes=current_record["size_bytes"],
            file_name=current_record["file_name"],
            number_of_chunks=total_number_of_chunks,
            blob_data=bytes(current_blob_data),
        )

  def _EnumerateCursor(
      self,
      cursor: sqlite3.Cursor,
      include_raw_data: bool = False,
      parse_key: bool = True,
      parse_value: bool = True,
      load_blobs: bool = True,
  ) -> Generator[ChromiumIndexedDBRecord, None, None]:
    """Yields ChromiumIndexedDBRecord records from a sqlite3 cursor.

    Args:
      cursor: the sqlite3 cursor.
      include_raw_data: whether to include the raw data.
      parse_key: whether to parse the key.
      parse_value: whether to parse the value.
      load_blobs: whether to load the record blobs.

    Yields:
      ChromiumIndexedDBRecord records.
    """
    for row in cursor:
      row_id = row[0]
      object_store_id = row[1]
      compression_type = definitions.DatabaseCompressionType(row[2])
      raw_key = row[3]
      raw_value = row[4]
      has_blobs = bool(row[5])

      key, value = None, None
      if parse_key and raw_key:
        key = record.SortableIDBKey.FromBytes(raw_data=raw_key, base_offset=0)

      if parse_value and raw_value:
        if compression_type == definitions.DatabaseCompressionType.UNCOMPRESSED:
          value = blink.V8ScriptValueDecoder.FromBytes(raw_value)
        elif compression_type == definitions.DatabaseCompressionType.ZSTD:
          value = blink.V8ScriptValueDecoder.FromBytes(
              zstd.decompress(raw_value)
          )
        elif compression_type == definitions.DatabaseCompressionType.SNAPPY:
          value = blink.V8ScriptValueDecoder.FromBytes(
              snappy.decompress(raw_value)
          )

      if load_blobs and raw_value is None:
        if not has_blobs:
          raise ValueError("Raw value is None but has_blobs is not set")
        value = []
        for blob in self.LoadBlobDataForRecordId(row_id):
          blob.blob_data = blink.V8ScriptValueDecoder.FromBytes(blob.blob_data)
          value.append(blob)

      yield ChromiumIndexedDBRecord(
          row_id=row_id,
          object_store_id=object_store_id,
          compression_type=compression_type,
          key=key,
          value=value,
          has_blobs=has_blobs,
          raw_key=raw_key if include_raw_data else None,
          raw_value=raw_value if include_raw_data else None,
      )

  def RecordsByObjectStoreId(
      self,
      object_store_id: int,
      include_raw_data: bool = False,
      parse_key: bool = True,
      parse_value: bool = True,
      load_blobs: bool = True,
  ) -> Generator[ChromiumIndexedDBRecord, None, None]:
    """Yields ChromiumIndexedDBRecord records for a given object store ID.

    Args:
      object_store_id: the object store ID.
      include_raw_data: whether to include the raw data.
      parse_key: whether to parse the key.
      parse_value: whether to parse the value.
      load_blobs: whether to load the record blobs.

    Yields:
      ChromiumIndexedDBRecord records.
    """
    with sqlite3.connect(f"{self._uri}?mode=ro", uri=True) as conn:
      conn.row_factory = sqlite3.Row
      cursor = conn.cursor()
      cursor.execute(definitions.SQL_RECORDS_BY_ID_QUERY, (object_store_id,))
      yield from self._EnumerateCursor(
          cursor, include_raw_data, parse_key, parse_value, load_blobs
      )

  def RecordsByObjectStoreName(
      self,
      object_store_name: str,
      include_raw_data: bool = False,
      parse_key: bool = True,
      parse_value: bool = True,
      load_blobs: bool = True,
  ) -> Generator[ChromiumIndexedDBRecord, None, None]:
    """Yields ChromiumIndexedDBRecord records for a given object store name.

    Args:
      object_store_name: the object store name.
      include_raw_data: whether to include the raw data.
      parse_key: whether to parse the key.
      parse_value: whether to parse the value.
      load_blobs: whether to load the record blobs.

    Yields:
      ChromiumIndexedDBRecord records.
    """
    with sqlite3.connect(f"{self._uri}?mode=ro", uri=True) as conn:
      conn.row_factory = sqlite3.Row
      cursor = conn.cursor()
      cursor.execute(
          definitions.SQL_RECORDS_BY_NAME_QUERY,
          (object_store_name.encode("utf-16-le"),),
      )
      yield from self._EnumerateCursor(
          cursor, include_raw_data, parse_key, parse_value, load_blobs
      )

  def Records(
      self,
      include_raw_data: bool = False,
      parse_key: bool = True,
      parse_value: bool = True,
      load_blobs: bool = True,
  ) -> Generator[ChromiumIndexedDBRecord, None, None]:
    """Yields ChromiumIndexedDBRecord records from all object stores.

    Args:
      include_raw_data: whether to include the raw data.
      parse_key: whether to parse the key.
      parse_value: whether to parse the value.
      load_blobs: whether to load the record blobs.

    Yields:
      ChromiumIndexedDBRecord records.
    """
    with sqlite3.connect(f"{self._uri}?mode=ro", uri=True) as conn:
      conn.row_factory = sqlite3.Row
      cursor = conn.cursor()
      cursor.execute(definitions.SQL_RECORDS_QUERY)
      yield from self._EnumerateCursor(
          cursor, include_raw_data, parse_key, parse_value, load_blobs
      )



# -*- coding: utf-8 -*-
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Safari IndexedDB records."""
import pathlib
import plistlib
import sqlite3
import sys
import traceback
from dataclasses import dataclass
from typing import Any, Generator, List, Optional

from dfindexeddb import errors
from dfindexeddb.indexeddb.safari import webkit


@dataclass
class SafariBlobInfo:
  """Safari IndexedDB blob info.

  Attributes:
    blob_url: the blob URL.
    file_name: the file name.
    file_path: the full path to the blob file.
    blob_data: the blob data.
  """

  blob_url: str
  file_name: str
  file_path: str
  blob_data: Optional[bytes] = None


@dataclass
class ObjectStoreInfo:
  """An ObjectStoreInfo.

  Attributes:
    id: the object store ID.
    name: the object store name.
    key_path: the object store key path.
    auto_inc: True if the object store uses auto incrementing IDs.
    database_name: the database name from the IDBDatabaseInfo table.
  """

  id: int
  name: str
  key_path: str
  auto_inc: bool
  database_name: str


@dataclass
class SafariIndexedDBRecord:
  """A Safari IndexedDBRecord.

  Attributes:
    key: the parsed key.
    value: the parsed value.
    object_store_id: the object store id.
    object_store_name: the object store name from the ObjectStoreInfo table.
    database_name: the IndexedDB database name from the IDBDatabaseInfo table.
    record_id: the record ID from the Record table.
    raw_key: the raw key.
    raw_value: the raw value.
  """

  key: Any
  value: Any
  object_store_id: int
  object_store_name: str
  database_name: str
  record_id: int
  raw_key: Optional[bytes] = None
  raw_value: Optional[bytes] = None
  blobs: Optional[List[SafariBlobInfo]] = None

  @property
  def is_key_filterable(self) -> bool:
    """True if the record key is filterable."""
    return True

  @property
  def is_value_filterable(self) -> bool:
    """True if the record value is filterable."""
    return True

  def MatchesKey(self, term: str) -> bool:
    """Returns True if the record key matches the filter term.

    Args:
      term: the filter term.
    """
    key_val = getattr(self.key, "value", self.key)
    return term in str(key_val)

  def MatchesValue(self, term: str) -> bool:
    """Returns True if the record value or blobs matches the filter term.

    Args:
      term: the filter term.
    """
    if term in str(self.value):
      return True

    if self.blobs:
      for blob_info in self.blobs:
        if blob_info.blob_data is not None and term in str(blob_info.blob_data):
          return True
    return False


class FileReader:
  """A reader for Safari IndexedDB sqlite3 files.

  Attributes:
    database_name: the database name.
    database_version: the database version.
    metadata_version: the metadata version.
    max_object_store_id: the maximum object store ID.
  """

  def __init__(self, filename: str):
    """Initializes the FileReader.

    Args:
      filename: the IndexedDB filename.
    """
    self.filename = filename
    self._uri = pathlib.Path(filename).resolve().as_uri()
    self.database_name = ""
    self.database_version = 0
    self.metadata_version = 0
    self.max_object_store_id = 0

    with sqlite3.connect(f"{self._uri}?mode=ro", uri=True) as conn:
      cursor = conn.execute(
          'SELECT value FROM IDBDatabaseInfo WHERE key = "DatabaseVersion"'
      )
      result = cursor.fetchone()
      if result:
        self.database_version = int(self._DecodeString(result[0]))

      cursor = conn.execute(
          'SELECT value FROM IDBDatabaseInfo WHERE key = "MetadataVersion"'
      )
      result = cursor.fetchone()
      if result:
        self.metadata_version = int(self._DecodeString(result[0]))

      cursor = conn.execute(
          'SELECT value FROM IDBDatabaseInfo WHERE key = "DatabaseName"'
      )
      result = cursor.fetchone()
      if result:
        self.database_name = self._DecodeString(result[0])

      cursor = conn.execute(
          'SELECT value FROM IDBDatabaseInfo WHERE key = "MaxObjectStoreID"'
      )
      result = cursor.fetchone()
      if result:
        self.max_object_store_id = int(self._DecodeString(result[0]))

  def _DecodeString(self, data: Any, vtype: Optional[str] = None) -> str:
    """Decodes a string from Safari IndexedDB metadata.

    Safari metadata strings are stored in SQLite as either:
    1. UTF-8 TEXT (returned as str or with vtype='text')
    2. UTF-16-LE BLOB (returned as bytes or with vtype='blob')

    Args:
      data: the string or bytes to decode.
      vtype: optional SQLite type ('text' or 'blob'). If not provided,
          isinstance checks are used.

    Returns:
      the decoded string.
    """
    if not data:
      return ""
    if isinstance(data, str):
      return data

    if isinstance(data, bytes):
      if vtype == "blob" or vtype is None:
        try:
          return data.decode("utf-16-le")
        except UnicodeDecodeError:
          pass
      return data.decode("utf-8", errors="replace")

    return str(data)

  def LoadBlobsForRecordId(self, record_id: int) -> List[SafariBlobInfo]:
    """Returns the SafariBlobInfo instances for the given record_id.

    Args:
      record_id: the record ID.

    Returns:
      a list of SafariBlobInfo instances.
    """
    blobs = []
    with sqlite3.connect(f"{self._uri}?mode=ro", uri=True) as conn:
      cursor = conn.execute(
          "SELECT r.blobURL, f.fileName "
          "FROM BlobRecords r "
          "JOIN BlobFiles f ON r.blobURL = f.blobURL "
          "WHERE r.objectStoreRow = ?",
          (record_id,),
      )
      for row in cursor:
        blob_url = self._DecodeString(row[0])
        file_name = self._DecodeString(row[1])

        # Check in .blobs subfolder first, then try the base directory
        blob_file_path = pathlib.Path(f"{self.filename}.blobs") / file_name
        if not blob_file_path.exists():
          blob_file_path = pathlib.Path(self.filename).parent / file_name

        blob_data = None
        if blob_file_path.exists():
          blob_data = blob_file_path.read_bytes()
        blobs.append(
            SafariBlobInfo(
                blob_url=blob_url,
                file_name=file_name,
                file_path=str(blob_file_path),
                blob_data=blob_data,
            )
        )
    return blobs

  def ObjectStores(self) -> Generator[ObjectStoreInfo, None, None]:
    """Returns the Object Store information from the IndexedDB database.

    Yields:
      ObjectStoreInfo instances.
    """
    with sqlite3.connect(f"{self._uri}?mode=ro", uri=True) as conn:
      cursor = conn.execute(
          "SELECT id, name, keypath, autoinc FROM ObjectStoreInfo"
      )
      results = cursor.fetchall()
      for result in results:
        key_path = plistlib.loads(result[2])
        yield ObjectStoreInfo(
            id=result[0],
            name=self._DecodeString(result[1]),
            key_path=key_path,
            auto_inc=result[3],
            database_name=self.database_name,
        )

  def _EnumerateCursor(
      self,
      cursor: sqlite3.Cursor,
      include_raw_data: bool = False,
      load_blobs: bool = True,
  ) -> Generator[SafariIndexedDBRecord, None, None]:
    """Yields SafariIndexedDBRecord records from a sqlite3 cursor.

    Args:
      cursor: the sqlite3 cursor.
      include_raw_data: whether to include the raw data.
      load_blobs: whether to load the record blobs.

    Yields:
      SafariIndexedDBRecord records.
    """
    for row in cursor:
      try:
        key = webkit.IDBKeyData.FromBytes(row[0]).data
      except (
          errors.ParserError,
          errors.DecoderError,
          NotImplementedError,
      ) as err:
        print(f"Error parsing IndexedDB key: {err}", file=sys.stderr)
        print(f"Traceback: {traceback.format_exc()}", file=sys.stderr)
        continue
      try:
        value = webkit.SerializedScriptValueDecoder.FromBytes(row[1])
      except (
          errors.ParserError,
          errors.DecoderError,
          NotImplementedError,
      ) as err:
        print(f"Error parsing IndexedDB value: {err}", file=sys.stderr)
        print(f"Traceback: {traceback.format_exc()}", file=sys.stderr)
        continue
      blobs = None
      if load_blobs:
        blobs = self.LoadBlobsForRecordId(row[5])
      yield SafariIndexedDBRecord(
          key=key,
          value=value,
          object_store_id=row[2],
          object_store_name=self._DecodeString(row[3], vtype=row[4]),
          database_name=self.database_name,
          record_id=row[5],
          raw_key=row[0] if include_raw_data else None,
          raw_value=row[1] if include_raw_data else None,
          blobs=blobs,
      )

  def RecordById(
      self,
      record_id: int,
      include_raw_data: bool = False,
      load_blobs: bool = True,
  ) -> Optional[SafariIndexedDBRecord]:
    """Returns an IndexedDBRecord for the given record_id.

    Returns:
      the IndexedDBRecord or None if the record_id does not exist in the
          database.
    """
    with sqlite3.connect(f"{self._uri}?mode=ro", uri=True) as conn:
      conn.text_factory = bytes
      cursor = conn.execute(
          "SELECT r.key, r.value, r.objectStoreID, o.name, typeof(o.name), "
          "r.recordID FROM "
          "Records r "
          "JOIN ObjectStoreInfo o ON r.objectStoreID == o.id "
          "WHERE r.recordID = ?",
          (record_id,),
      )
      try:
        return next(self._EnumerateCursor(cursor, include_raw_data, load_blobs))
      except StopIteration:
        return None

  def RecordsByObjectStoreName(
      self,
      name: str,
      include_raw_data: bool = False,
      load_blobs: bool = True,
  ) -> Generator[SafariIndexedDBRecord, None, None]:
    """Returns IndexedDBRecords for the given ObjectStore name.

    Yields:
      IndexedDBRecord instances.
    """
    with sqlite3.connect(f"{self._uri}?mode=ro", uri=True) as conn:
      conn.text_factory = bytes
      cursor = conn.execute(
          "SELECT r.key, r.value, r.objectStoreID, o.name, typeof(o.name), "
          "r.recordID FROM "
          "Records r "
          "JOIN ObjectStoreInfo o ON r.objectStoreID == o.id "
          "WHERE o.name = ?",
          (name,),
      )
      yield from self._EnumerateCursor(cursor, include_raw_data, load_blobs)

  def RecordsByObjectStoreId(
      self,
      object_store_id: int,
      include_raw_data: bool = False,
      load_blobs: bool = True,
  ) -> Generator[SafariIndexedDBRecord, None, None]:
    """Returns IndexedDBRecords for the given ObjectStore id.

    Yields:
      IndexedDBRecord instances.
    """
    with sqlite3.connect(f"{self._uri}?mode=ro", uri=True) as conn:
      conn.text_factory = bytes
      cursor = conn.execute(
          "SELECT r.key, r.value, r.objectStoreID, o.name, typeof(o.name), "
          "r.recordID "
          "FROM Records r "
          "JOIN ObjectStoreInfo o ON r.objectStoreID == o.id "
          "WHERE o.id = ?",
          (object_store_id,),
      )
      yield from self._EnumerateCursor(cursor, include_raw_data, load_blobs)

  def Records(
      self, include_raw_data: bool = False, load_blobs: bool = True
  ) -> Generator[SafariIndexedDBRecord, None, None]:
    """Returns all the IndexedDBRecords."""
    with sqlite3.connect(f"{self._uri}?mode=ro", uri=True) as conn:
      conn.text_factory = bytes
      cursor = conn.execute(
          "SELECT r.key, r.value, r.objectStoreID, o.name, typeof(o.name), "
          "r.recordID "
          "FROM Records r "
          "JOIN ObjectStoreInfo o ON r.objectStoreID == o.id"
      )
      yield from self._EnumerateCursor(cursor, include_raw_data, load_blobs)



# -*- coding: utf-8 -*-
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helper/utility classes for LevelDB."""
from __future__ import annotations

import io
import struct
from typing import BinaryIO, Tuple, Type, TypeVar

from dfindexeddb import errors, utils


_CHUNK_SIZE = 8
_MARKER = _CHUNK_SIZE + 1
_EMPTY_BINARY_SENTINEL = 0
_SIGN_BIT = 1 << 63
_SENTINEL = 0
_TWO_BYTE_ENCODING_INDICATOR = 0x80
_THREE_BYTE_ENCODING_INDICATOR = 0xFF


class LevelDBDecoder(utils.StreamDecoder):
  """A helper class to decode data types from LevelDB files."""

  def DecodeBool(self) -> Tuple[int, bool]:
    """Returns a Tuple of the offset of decoding and the bool value."""
    offset, buffer = self.ReadBytes(1)
    return offset, buffer[0] is not None

  def DecodeString(self) -> Tuple[int, str]:
    """Returns a tuple of the offset of decoding and the string value.

    Raises:
      errors.DecoderError: when the parsed string buffer is not even (i.e.
          cannot be decoded as a UTF-16-BE string.
    """
    offset = self.stream.tell()
    buffer = self.stream.read()
    if len(buffer) % 2:
      raise errors.DecoderError(
          f"Odd number of bytes encountered at offset {offset}"
      )
    return offset, buffer.decode("utf-16-be")

  def DecodeLengthPrefixedSlice(self) -> Tuple[int, bytes]:
    """Returns a tuple of the offset of decoding and the byte 'slice'."""
    offset, num_bytes = self.DecodeUint32Varint()
    _, blob = self.ReadBytes(num_bytes)
    return offset, blob

  def DecodeBlobWithLength(self) -> Tuple[int, bytes]:
    """Returns a tuple of a the offset of decoding and the binary blob."""
    offset, num_bytes = self.DecodeUint64Varint()
    _, blob = self.ReadBytes(num_bytes)
    return offset, blob

  def DecodeStringWithLength(
      self, encoding: str = "utf-16-be"
  ) -> Tuple[int, str]:
    """Returns a tuple of the offset of decoding and the string value."""
    offset, length = self.DecodeUint64Varint()
    _, buffer = self.ReadBytes(length * 2)
    return offset, buffer.decode(encoding=encoding)

  def DecodeSortableBinary(self) -> Tuple[int, bytes]:
    """Decodes a sortable binary from the binary stream.

    Returns:
      a tuple of the offset and the decoded bytes.

    Raises:
      errors.ParserError: if an invalid tag is encountered.
    """
    output = bytearray()

    offset, first = self.PeekBytes(1)
    if first[0] == _EMPTY_BINARY_SENTINEL:
      self.ReadBytes(1)
      return offset, b""

    while True:
      _, marker_or_sentinel = self.DecodeUint8()
      if marker_or_sentinel == _MARKER:
        _, chunk = self.ReadBytes(_CHUNK_SIZE)
        output.extend(chunk)
        continue
      if 1 <= marker_or_sentinel <= _CHUNK_SIZE:
        payload_len = marker_or_sentinel
        padding_len = _CHUNK_SIZE - payload_len
        if padding_len > 0:
          return offset, bytes(output[:-padding_len])
        return offset, bytes(output)
      raise errors.ParserError(
          f"Invalid marker or sentinel {marker_or_sentinel} in sortable binary"
      )

  def DecodeSortableDouble(self) -> Tuple[int, float]:
    """Decodes a sortable double-precision float from the binary stream.

    Returns:
      a tuple of the offset and the decoded double.
    """
    offset, host_bits = self.DecodeInt(8, byte_order="big", signed=False)
    if host_bits & _SIGN_BIT:
      host_bits ^= _SIGN_BIT
    else:
      host_bits ^= 0xFFFFFFFFFFFFFFFF
    blob = host_bits.to_bytes(8, byteorder="big")
    return offset, struct.unpack(">d", blob)[0]

  def DecodeSortableString(self) -> Tuple[int, str]:
    """Decodes a sortable string from the binary stream.

    Returns:
      a tuple of the offset and the decoded string.

    Raises:
      errors.ParserError: if an invalid byte is encountered.
    """
    output = []
    offset = self.stream.tell()
    while True:
      _, first = self.DecodeUint8()
      if first == _SENTINEL:
        break
      if (first & 0x80) == 0:
        output.append(chr((first & 0x7F) - 1))
      elif (first & 0xC0) == _TWO_BYTE_ENCODING_INDICATOR:
        _, second = self.DecodeUint8()
        output.append(chr(((first & 0x3F) << 8) | second))
      elif first == _THREE_BYTE_ENCODING_INDICATOR:
        _, high = self.DecodeUint8()
        _, low = self.DecodeUint8()
        output.append(chr((high << 8) | low))
      else:
        raise errors.ParserError(f"Invalid byte {first} in sortable string")
    return offset, "".join(output)


T = TypeVar("T")


class FromDecoderMixin:
  """A mixin for parsing dataclass attributes using a LevelDBDecoder."""

  @classmethod
  def FromDecoder(
      cls: Type[T], decoder: LevelDBDecoder, base_offset: int = 0
  ) -> T:
    """Decodes a class type from the current position of a LevelDBDecoder.

    Args:
      decoder: the LevelDBDecoder.
      base_offset: the base offset.

    Returns:
      The class instance.

    Raises:
      NotImplementedError if the child class does not implement this method.
    """
    raise NotImplementedError

  @classmethod
  def FromStream(cls: Type[T], stream: BinaryIO, base_offset: int = 0) -> T:
    """Decodes a class type from the current position of a binary stream.

    Args:
      stream: the binary stream.
      base_offset: the base offset of the binary stream.

    Returns:
      The class instance.
    """
    decoder = LevelDBDecoder(stream)
    return cls.FromDecoder(  # type: ignore[attr-defined,no-any-return]
        decoder=decoder, base_offset=base_offset
    )

  @classmethod
  def FromBytes(
      cls: Type[T], raw_data: bytes | bytearray, base_offset: int = 0
  ) -> T:
    """Parses a class type from raw bytes.

    Args:
      raw_data: the raw data.
      base_offset: the base offset of the raw data.

    Returns:
      The class instance.
    """
    stream = io.BytesIO(raw_data)
    return cls.FromStream(  # type: ignore[attr-defined,no-any-return]
        stream=stream, base_offset=base_offset
    )
